unit SynHighlighterObr;

interface

uses
  SysUtils, LCLProc, Dialogs, Classes, Graphics, SynEdit,
  SynEditHighlighter, SynEditHighlighterFoldBase, SynGutterCodeFolding;

type

  TtkTokenKind = (tkAsm, tkComment, tkIdentifier, tkKey, tkNull, tkNumber,
    tkSpace, tkAnsiString, tkString, tkSymbol, tkSpecial, tkUnknown, tkSet,
    tkTypes, tkSystem, tkSysmod, tkDirective);

  TRangeState = (
      rsAnsi,
      rsBlock,
      rsImport,
      rsEqualOrColon,
      rsProc,
      rsLocal,
      rsBracket
    );
  TRangeStates = set of TRangeState;

  TObrCodeFoldBlockType = (
    cfbtBeginEnd,
    cfbtProcedure,
    cfbtComment
  );

  TSynObrSyn = class(TSynCustomFoldHighlighter)
  private

    fLineStr             : string;
    fTokenName           : string;

    fLine                : PChar;
    fLineLen             : integer;
    Run                  : LongInt;
    fStringLen           : Integer;
    fTokenPos            : Integer;
    fTokenID             : TtkTokenKind;

    fLineNumber          : integer;
    fRange               : TRangeStates;

    fStringAttri         : TSynHighlighterAttributes;
    fAnsiStringAttri     : TSynHighlighterAttributes;
    fNumberAttri         : TSynHighlighterAttributes;
    fKeyAttri            : TSynHighlighterAttributes;
    fSymbolAttri         : TSynHighlighterAttributes;
    fSetAttri            : TSynHighlighterAttributes;
    fCommentAttri        : TSynHighlighterAttributes;
    fDirectiveAttri      : TSynHighlighterAttributes;
    fIdentifierAttri     : TSynHighlighterAttributes;
    fSpaceAttri          : TSynHighlighterAttributes;
    fSpecialAttri        : TSynHighlighterAttributes;

    fTypAttri            : TSynHighlighterAttributes;
    fSysAttri            : TSynHighlighterAttributes;
    fModAttri            : TSynHighlighterAttributes;

    procedure   CRProc;
    procedure   LFProc;
    procedure   NullProc;
    procedure   UnknownProc;
    procedure   SpaceProc;

    procedure   RoundOpenProc;
    procedure   SymbolProc;

    procedure   StringProc;
    procedure   NumberProc;

    procedure   AnsiProc;
    procedure   IdentProc;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function    GetEol: Boolean; override;
    function    GetRange: Pointer; override;
    function    GetToken: string; override;
    procedure   GetTokenEx(out TokenStart: PChar; out TokenLength: integer); override;
    function    GetTokenAttribute: TSynHighlighterAttributes; override;
    function    GetTokenID: TtkTokenKind;
    function    GetTokenKind: integer; override;
    function    GetTokenPos: Integer; override;
    function    GetDefaultAttribute(Index: integer): TSynHighlighterAttributes; override;

    procedure   Next; override;

    procedure   ResetRange; override;
    procedure   SetLine({$IFDEF FPC}const {$ENDIF}NewValue: string; LineNumber: Integer); override;
    procedure   SetRange(Value: Pointer); override;

  published
    property SetAttri         : TSynHighlighterAttributes read fSetAttri        write fSetAttri;
    property CommentAttri     : TSynHighlighterAttributes read fCommentAttri    write fCommentAttri;
    property DirectiveAttri   : TSynHighlighterAttributes read fDirectiveAttri  write fDirectiveAttri;
    property IdentifierAttri  : TSynHighlighterAttributes read fIdentifierAttri write fIdentifierAttri;
    property KeyAttri         : TSynHighlighterAttributes read fKeyAttri        write fKeyAttri;
    property NumberAttri      : TSynHighlighterAttributes read fNumberAttri     write fNumberAttri;
    property SpaceAttri       : TSynHighlighterAttributes read fSpaceAttri      write fSpaceAttri;
    property AnsiStringAttri  : TSynHighlighterAttributes read fAnsiStringAttri write fAnsiStringAttri;
    property StringAttri      : TSynHighlighterAttributes read fStringAttri     write fStringAttri;
    property SymbolAttri      : TSynHighlighterAttributes read fSymbolAttri     write fSymbolAttri;
    property SpecialAttri     : TSynHighlighterAttributes read fSpecialAttri    write fSpecialAttri;
    property TypAttri         : TSynHighlighterAttributes read fTypAttri        write fTypAttri;
    property SysAttri         : TSynHighlighterAttributes read fSysAttri        write fSysAttri;
    property ModAttri         : TSynHighlighterAttributes read fModAttri        write fModAttri;
  end;

implementation

var
  KeywordsList: TStringList;
  Identifiers: array[char] of Boolean;
  IsIntegerChar: array[char] of Boolean;
  IsNumberChar: array[char] of Boolean;
  IsSpaceChar: array[char] of Boolean;

const

  TokenKey : array [0..48] of string[10]=(
    'IF','DO','OF','OR','TO','IN','IS','BY',
    'END','NIL','VAR','DIV','MOD','FOR',
    'INC','DEC','INCL','EXCL','NEW','ASSERT',
    'PACK','UNPK','DISPOSE','ABS','ODD','LEN',
    'LSL','ASR','ROR','FLOOR','FLT','ORD','CHR',
    'ELSE','THEN','TYPE','CASE',
    'ELSIF','ARRAY','BEGIN','CONST','UNTIL','WHILE',
    'RECORD','REPEAT','MODULE','POINTER','ADDRESS','PROCEDURE'
  );

  TypesKey : array [0..13] of string[10]=(
    'INTEGER','INT8','INT16','INT32','BYTE','CARD16','CARD32',
    'BOOLEAN','SET','CHAR','CHAR8','REAL','LONGINT','LONGREAL'
  );

  SystemKey : array [0..6] of string[10]=(
    'TRUE','FALSE','IMPORT','RETURN','EXIT','UNION','EXTENSIBLE'
  );

  SystemMod : array [0..11] of string[15]=(
    'SYSTEM','GET','PUT','COPY','LoadLibraryW','GetProcAddress',
    'ADR','SIZE','BIT','VAL','ADR2','STRADR'
  );

constructor TSynObrSyn.Create(AOwner: TComponent);
var
  i: char;
  n: integer;
begin
  inherited Create(AOwner);

  for i := #0 to #255 do begin
    Identifiers[i] :=(i in ['_', '0'..'9', 'a'..'z', 'A'..'Z']);
    IsIntegerChar[i]:=(i in ['0'..'9', 'A'..'F']);
    IsNumberChar[i]:=(i in ['0'..'9']);
    IsSpaceChar[i]:=(i in [#1..#9, #11, #12, #14..#32]);
  end;

  if KeywordsList = nil then begin
    KeywordsList := TStringList.Create;
    for n:=0 to High(TokenKey)      do KeywordsList.AddObject(TokenKey[n],     TObject(tkKey));
    for n:=0 to High(TypesKey)      do KeywordsList.AddObject(TypesKey[n],     TObject(tkTypes));
    for n:=0 to High(SystemKey)     do KeywordsList.AddObject(SystemKey[n],    TObject(tkSystem));
    for n:=0 to High(SystemMod)     do KeywordsList.AddObject(SystemMod[n],    TObject(tkSysmod));
    KeywordsList.Sort;
    KeywordsList.Sorted := FALSE;
  end;

  fSetAttri := TSynHighlighterAttributes.Create('Set', 'Set');
  fSetAttri.Foreground := $00996600;
  fSetAttri.Style     := [fsItalic];
  AddAttribute(fSetAttri);

  fCommentAttri := TSynHighlighterAttributes.Create('Comment', 'Comment');
  fCommentAttri.Style:= [fsItalic];
  fCommentAttri.Foreground := $00888888;
  AddAttribute(fCommentAttri);

  fDirectiveAttri := TSynHighlighterAttributes.Create('Directive', 'Directive');
  fDirectiveAttri.Style:= [fsBold];
  fDirectiveAttri.Foreground := $00888800;
  AddAttribute(fDirectiveAttri);

  fKeyAttri := TSynHighlighterAttributes.Create('Reserved', 'Reserved');
  fKeyAttri.Style     := [fsBold];
  fKeyAttri.Foreground := $00880000;
  AddAttribute(fKeyAttri);

  fTypAttri           := TSynHighlighterAttributes.Create('Type','Type');
  fTypAttri.Style     := [fsBold];
  fTypAttri.Foreground := $00008800;
  AddAttribute(fTypAttri);

  fSysAttri           := TSynHighlighterAttributes.Create('System','System');
  fSysAttri.Style     := [fsBold];
  fSysAttri.Foreground := $000000BB;
  AddAttribute(fSysAttri);

  fModAttri           := TSynHighlighterAttributes.Create('Module','Module');
  fModAttri.Style     := [fsBold];
  fModAttri.Foreground := $00000000;
  AddAttribute(fModAttri);

  fNumberAttri := TSynHighlighterAttributes.Create('Number', 'Number');
  //fNumberAttri.Style     := [fsBold];
  fNumberAttri.Foreground := $00996600;
  AddAttribute(fNumberAttri);

  fSpaceAttri := TSynHighlighterAttributes.Create('Space', 'Space');
  AddAttribute(fSpaceAttri);

  fStringAttri := TSynHighlighterAttributes.Create('String', 'String');
  fStringAttri.Foreground:=$00000088;
  fStringAttri.Style:= [fsBold];
  AddAttribute(fStringAttri);

  fAnsiStringAttri := TSynHighlighterAttributes.Create('AString', 'AString');
  fAnsiStringAttri.Foreground:=$005500CC;
  fAnsiStringAttri.Style:= [fsBold];
  AddAttribute(fAnsiStringAttri);

  fSymbolAttri := TSynHighlighterAttributes.Create('Symbol', 'Symbol');
  //fSymbolAttri.Style     := [fsBold];
  fSymbolAttri.Foreground := $00996600;
  AddAttribute(fSymbolAttri);

  fSpecialAttri := TSynHighlighterAttributes.Create('Special', 'Special');
  fSpecialAttri.Style     := [fsBold];
  fSpecialAttri.Foreground := $00002288;
  AddAttribute(fSpecialAttri);

  fRange := [];

end;

function TSynObrSyn.GetTokenAttribute: TSynHighlighterAttributes;
begin
  case GetTokenID of
    tkSet         : Result := fSetAttri;
    tkComment     : Result := fCommentAttri;
    tkDirective   : Result := fDirectiveAttri;
    tkIdentifier  : Result := fIdentifierAttri;
    tkKey         : Result := fKeyAttri;
    tkTypes       : Result := fTypAttri;
    tkSystem      : Result := fSysAttri;
    tkSysmod      : Result := fModAttri;
    tkNumber      : Result := fNumberAttri;
    tkSpace       : Result := fSpaceAttri;
    tkAnsiString  : Result := fAnsiStringAttri;
    tkString      : Result := fStringAttri;
    tkSymbol      : Result := fSymbolAttri;
    tkSpecial     : Result := fSpecialAttri;
  else
    Result := nil;
  end;
end;

function TSynObrSyn.GetDefaultAttribute(Index: integer): TSynHighlighterAttributes;
begin
  if index>=0 then Result:=nil;
end;

destructor TSynObrSyn.Destroy;
begin
  inherited Destroy;
end;

procedure TSynObrSyn.SetLine(const NewValue: string; LineNumber:Integer);
begin
  fLineStr := NewValue;
  fLineLen:=length(fLineStr);
  fLine:=PChar(Pointer(fLineStr));
  fLineNumber := LineNumber;
  Run := 0;
  Inherited SetLine(NewValue,LineNumber);
  Next;
end; { SetLine }

function TSynObrSyn.GetEol: Boolean;
begin
  Result := (fTokenID = tkNull) and (Run >= fLineLen);
end;

function TSynObrSyn.GetToken: string;
var
  Len: LongInt;
begin
  Len := Run - fTokenPos;
  SetLength(Result,Len);
  if Len>0 then System.Move(fLine[fTokenPos],Result[1],Len);
end;

procedure TSynObrSyn.GetTokenEx(out TokenStart: PChar; out TokenLength: integer);
begin
  TokenLength:=Run-fTokenPos;
  if TokenLength>0 then begin
    TokenStart:=@fLine[fTokenPos];
  end else begin
    TokenStart:=nil;
  end;
end;

function TSynObrSyn.GetTokenKind: integer;
begin
  Result := Ord(GetTokenID);
end;

function TSynObrSyn.GetTokenID: TtkTokenKind;
begin
  Result := fTokenId;
end;

function TSynObrSyn.GetTokenPos: Integer;
begin
  Result := fTokenPos;
end;

function TSynObrSyn.GetRange: Pointer;
begin
  CodeFoldRange.RangeType:=Pointer(PtrUInt(Integer(fRange)));
  Result := inherited GetRange;
end;

procedure TSynObrSyn.SetRange(Value: Pointer);
begin
  inherited SetRange(Value);
  fRange := TRangeStates(Integer(PtrUInt(CodeFoldRange.RangeType)));
end;

procedure TSynObrSyn.ResetRange;
begin
  fRange := [];
  inherited ResetRange;
end;

(*########################################*)

procedure TSynObrSyn.CRProc;
begin
  fTokenID := tkSpace;
  inc(Run);
  if fLine[Run] = #10 then inc(Run);
end;

procedure TSynObrSyn.LFProc;
begin
  fTokenID := tkSpace;
  inc(Run);
end;

procedure TSynObrSyn.NullProc;
begin
  fTokenID := tkNull;
  if Run<fLineLen then inc(Run);
end;

procedure TSynObrSyn.UnknownProc;
begin
  inc(Run);
  while (fLine[Run] in [#128..#191]) do inc(Run);
  fTokenID := tkUnknown;
end;

procedure TSynObrSyn.SpaceProc;
begin
  inc(Run);
  fTokenID := tkSpace;
  while IsSpaceChar[FLine[Run]] do inc(Run);
end;

procedure TSynObrSyn.StringProc;
var
  c: char;
begin
  fTokenID := tkString;
  c:=fLine[Run];
  Inc(Run);
  while (not (fLine[Run] in [#0, #10, #13])) do begin
    if fLine[Run] = c then begin
      Inc(Run);
      if (fLine[Run] <> c) then break;
    end;
    Inc(Run);
  end;
  if (fLine[Run]='@') then begin
    Inc(Run);
    fTokenID := tkAnsiString;
  end;
end;

procedure TSynObrSyn.NumberProc;
var
  i,HexFlag,DecFlag,FloatFlag: integer;
  Word: string;
begin
  i:=Run;
  fTokenID := tkUnknown;

  HexFlag:=0;
  DecFlag:=0;
  FloatFlag:=0;

  while Identifiers[fLine[Run]] do inc(Run);

  Word := copy(fLine,i+1,Run-i);

  for i:=1 to length(Word) do begin
    if not IsIntegerChar[Word[i]] then inc(HexFlag);
    if not IsNumberChar[Word[i]]  then
      if (Word[i] in ['E','D']) and (FloatFlag=0) then
        inc(FloatFlag)
      else
        inc(DecFlag);
  end;

  if HexFlag>1 then exit;

  fTokenID := tkNumber;
  if DecFlag=0 then exit;

  case Word[i] of
    'H': fTokenID := tkNumber;
    'X': if (i<=4) then fTokenID := tkString else fTokenID := tkUnknown;
  else
    fTokenID := tkUnknown;
  end;

end;

procedure TSynObrSyn.RoundOpenProc;
begin
  Inc(Run);
  if Run>=fLineLen then begin
    fTokenID:=tkSymbol;
    exit;
  end;
  case fLine[Run] of
    '*':
      begin
        fRange := fRange + [rsAnsi];
        fTokenID := tkComment;
        fStringLen := 2;
        Dec(Run);
        StartCodeFoldBlock(Pointer(PtrInt(cfbtComment)));
        Inc(Run,2);
        if fLine[Run]='$' then begin
          fStringLen := 3;
          fTokenID := tkDirective;
          Inc(Run);
        end;
        if not (fLine[Run] in [#0, #10, #13]) then AnsiProc;
      end;
  else
    fTokenID := tkSymbol;
  end;
end;

procedure TSynObrSyn.AnsiProc;
begin
  if fTokenID<>tkDirective then fTokenID := tkComment;
  repeat
    if fLine[Run]=#0 then
      break
    else if (fLine[Run] = '*') and (fLine[Run + 1] = ')') then begin
      Inc(Run, 2);
      EndCodeFoldBlock;
      if TopCodeFoldBlockType <> Pointer(PtrInt(cfbtComment)) then
        fRange := fRange - [rsAnsi];
      break;
    end else if (fLine[Run] = '(') and (fLine[Run + 1] = '*') then begin
      fStringLen := 2;
      StartCodeFoldBlock(Pointer(PtrInt(cfbtComment)));
      Inc(Run,2);
    end else
      Inc(Run);
  until (Run>=fLineLen) or (fLine[Run] in [#0, #10, #13]);
  if fTokenID=tkDirective then fRange := fRange - [rsAnsi];
end;

procedure TSynObrSyn.SymbolProc;
var
  c: char;
  n: integer;
begin
  c:=fLine[Run];

  if c='(' then begin
    RoundOpenProc;
    exit;
  end;

  inc(Run);
  fTokenID := tkSymbol;

  case c of
    '{': fRange := fRange + [rsBracket];
    '}': fRange := fRange - [rsBracket];
    '<': if fLine[Run] in ['=']      then begin inc(Run); fStringLen := 2; end;
    '>': if fLine[Run] in ['=', '<'] then begin inc(Run); fStringLen := 2; end;
    ':': if fLine[Run] in ['=']      then begin inc(Run); fStringLen := 2; end;
    '|': fRange:=fRange-[rsEqualOrColon];
  end;

  case c of
    '=',':': fRange := fRange + [rsEqualOrColon];
    ';'    :
             begin
               fRange:=fRange-[rsImport,rsEqualOrColon];
               if rsProc in fRange then fRange:=fRange-[rsProc]+[rsLocal];
             end;
    ')'    : ;
  end;

end;

procedure TSynObrSyn.IdentProc;
var
  i,n : integer;
  Word: string;
  Flag: boolean;

  function CheckProc(i: integer): boolean;
  begin
    Result:=true;
    while ((fLine[i]<>';') and (i<=fLineLen)) do begin
      if fLine[i]='(' then exit;
      if fLine[i] IN [',','^'] then begin Result:=false; exit; end;
      inc(i);
    end;
  end;

begin
  i:=Run;
  fTokenID := tkIdentifier;

  while Identifiers[fLine[Run]] do inc(Run);

  Word := copy(fLine,i+1,Run-i);
  fTokenName:= Word;

  Flag:=true;

  for n:=0 to KeywordsList.Count-1 do
    if Word[1] = KeywordsList[n][1] then
    if Word = KeywordsList[n] then begin

  //if KeywordsList.Find(Word, n) then
    //if Word=KeywordsList.Strings[n] then begin

      fTokenID := TtkTokenKind(PtrUInt(KeywordsList.Objects[n]));

      if Word<>'SYSTEM' then fRange:=fRange-[rsImport];

      case Word[1] of
        'B': if Word='BEGIN'  then fRange:=fRange+[rsBlock]-[rsLocal];
        'I': if Word='IMPORT' then fRange:=fRange+[rsImport];
        'R': if Word='RETURN' then fRange:=fRange+[rsBlock]-[rsLocal];
        'S': if Word='SYSTEM' then if rsImport in fRange then fTokenID := tkSpecial;
      end;

      case Word[1] of
        'C': if  Word='CASE'                    then StartCodeFoldBlock(nil);
        'F': if  Word='FOR'                     then StartCodeFoldBlock(nil);
        'I': if  Word='IF'                      then StartCodeFoldBlock(nil);
        'L': if  Word='LOOP'                    then StartCodeFoldBlock(nil);
        'R': if  Word='RECORD'                  then StartCodeFoldBlock(nil);
        'W': if  Word='WHILE'                   then StartCodeFoldBlock(nil);
        'P': if  Word='PROCEDURE'               then if not (rsEqualOrColon in fRange) and (CheckProc(Run)) then begin
                                                    StartCodeFoldBlock(Pointer(PtrInt(cfbtProcedure)));
                                                    fRange:=fRange+[rsProc];
                                                end;
      end;

      case Word[1] of
        'R': if  Word='RECORD'                  then fRange:=fRange-[rsEqualOrColon];
        'P': if  Word='POINTER'                 then fRange:=fRange-[rsEqualOrColon];
        'E': if  Word='END'                     then if TopCodeFoldBlockType=Pointer(PtrInt(cfbtProcedure)) then fRange:=fRange-[rsBlock];
      end;

      case Word[1] of
        'R': if  Word='REPEAT'                  then StartCodeFoldBlock(nil);
        'U': if  Word='UNTIL'                   then EndCodeFoldBlock;
        'E': if  Word='END'                     then EndCodeFoldBlock;
      end;

      exit;
    end;

  if rsImport in fRange then fTokenID := tkSpecial;

  if (fRange*[rsEqualOrColon,rsBlock,rsLocal]=[]) then begin
    n:=Run;
    while fLine[n] in [#32] do inc(n);
    if fLine[n]='*' then begin
      inc(Run,n-Run+1);
      fTokenID := tkSysmod;
    end;
  end;

end;

procedure TSynObrSyn.Next;
begin

  fTokenPos := Run;

  if Run>=fLineLen then begin
    NullProc;
    exit;
  end;

  case fLine[Run] of
     #0                         : @NullProc;
    #10                         : @LFProc;
    #13                         : @CRProc;
    else

    if rsAnsi in fRange then begin
      AnsiProc;
    end else begin

      case fLine[Run] of
        #1..#9, #11, #12, #14..#32  : @SpaceProc;
        'A'..'Z', 'a'..'z', '_'     : @IdentProc;
        '''','"'                    : @StringProc;
        '0'..'9'                    : @NumberProc;
        '{','}', '!', '%', '&','|',
        '('..'/', ':'..'@','#',
        '['..'^', '`', '~'          : @SymbolProc;
      else
        @UnknownProc;
      end;

    end;

  end;

end;

initialization
  RegisterPlaceableHighlighter(TSynObrSyn);

finalization
  FreeAndNil(KeywordsList);

end.


