unit DockManager;

{$mode objfpc}{$H+}

interface

uses
  Classes, Controls, StdCtrls, SynEdit,
  Xscontrols, Graphics, Dialogs, SysUtils,
  SynGutter, SynGutterCodeFolding, SynGutterLineNumber,
  SynHighlighterObr, SynEditKeyCmds, SynEditMarkupHighAll,
  SynEditMouseCmds, SynEditHighlighter;

const
  sep = DirectorySeparator;

type

  { TXSDockManager }

  TXSDockManager = class
  private
    feditor     : TSynEdit;
    fdefname    : string;
    fdefext     : string;
    ffilter     : string;
    fopendialog : TOpenDialog;
    fsavedialog : TSaveDialog;
    fpagecontrol: TXSpagecontrol;
    procedure   SetEditor(const AValue: TSynEdit);
    function    GetEditorByNum(num,idx: integer): TSynEdit;
    function    GetEditor:TSynEdit;
    procedure   InvalidateParent(p: TControl);
    procedure   SwapCommand(idx1,idx2: TSynEditorMouseCommand);
    procedure   EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorMouseLink(Sender: TObject; X, Y: Integer; var AllowMouseLink: Boolean);
    procedure   EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
    procedure   EditorOptions(ed: TSynEdit; tab: TXSScrollBox);
    function    GetTabByNum(num,idx: integer): TXSScrollBox;
 public
    wordatxy    : string;
    constructor Create(Page: TXSpagecontrol);
    destructor  Destroy; override;
    procedure   UpdateText;
    procedure   PageChange(Sender: TObject);
    procedure   EditChange(Sender: TObject);
    procedure   ScrollChange(Sender: TObject);
    procedure   EditPaint(Sender: TObject; ACanvas: TCanvas);
    function    AllModified: boolean;
    procedure   PageClose(Sender: TObject; idx: integer);
    procedure   MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

    procedure   NewPage;
    procedure   OpenPage;
    procedure   ClosePage;

    function    SaveFileAsByIdx(idx: integer): Boolean;
    procedure   SaveFileAs;
    function    SaveFileByIdx(idx: integer): Boolean;
    procedure   SaveFile;
    procedure   SaveAll;
    function    CloseFileByIdx(idx: integer): Boolean;
    procedure   CloseFile;
    function    CloseAll: Boolean;
    function    SaveCloseQuery: Boolean;

    procedure   Cut;
    procedure   Copy;
    procedure   Paste;
    procedure   Delete;
    procedure   Undo;
    procedure   Redo;
    procedure   Find;
    procedure   FindNext;
    procedure   FindPrev;
    procedure   Unindent;
    procedure   Indent;
    procedure   Build;
    procedure   Run;
    procedure   Pause;
    procedure   Stop;
    procedure   SelectAll;
    procedure   DeleteSelection;
    procedure   Upper;
    procedure   Lower;
    procedure   Swap;
    procedure   Title;

  published
    property    Pagecontrol: TXSpagecontrol read fpagecontrol write fpagecontrol;
    property    Editor     : TSynEdit read GetEditor write SetEditor;
    property    DefaultName: string   read fdefname write fdefname;
    property    DefaultExt : string   read fdefext write fdefext;
    property    FileFilter : string   read ffilter write ffilter;
  end;

implementation

uses
  main;

{ TXSDockManager }

constructor TXSDockManager.Create(Page: TXSpagecontrol);
begin
  inherited Create;
  fdefname := 'Untitled';
  fdefext  := '.txt';
  ffilter  := 'Text files (*.txt;*.ini;*.inc)|*.txt;*.ini;*.inc'
             +'|All files (*.*)|*.*';

  fopendialog         := TOpenDialog.Create(nil);
  fopendialog.Options := fopendialog.Options + [ofAllowMultiSelect];
  fsavedialog         := TSaveDialog.Create(nil);
  fpagecontrol        := Page;
  fpagecontrol.OnChange    := @PageChange;
  fpagecontrol.OnMouseDown := @MouseDown;
  fpagecontrol.OnPageClose := @PageClose;

  fpagecontrol.AddText('1:1',70);
  fpagecontrol.AddText('ANSI',40);
  fpagecontrol.AddText('',55);
  fpagecontrol.AddText('',250);
end;

destructor TXSDockManager.Destroy;
begin
  FreeAndNil(fopendialog);
  FreeAndNil(fsavedialog);
end;

procedure TXSDockManager.UpdateText;
begin
  if fpagecontrol.pagelist.count=0 then exit;
  fpagecontrol.Text[0] := IntToStr(Editor.CaretX)+':'+IntToStr(Editor.CaretY);
  fpagecontrol.Text[2] := '';
  if Editor.Modified then fpagecontrol.Text[2] := 'Modified';
  fpagecontrol.Text[3] := GetTabByNum(pagecontrol.activepage,0).Path;
end;

procedure TXSDockManager.SetEditor(const AValue: TSynEdit);
begin
  if AValue<>feditor then
    feditor := AValue;
end;

function TXSDockManager.GetTabByNum(num, idx: integer): TXSScrollBox;
begin
  result := TXSScrollBox(TStringList(fpagecontrol.pagelist.objects[num]).objects[idx]);
end;

function TXSDockManager.GetEditorByNum(num,idx: integer): TSynEdit;
var
  tab: TXSScrollbox;
  i: integer;
  f: boolean;
begin
  result := NIL;
  tab := GetTabByNum(num,0);
  i:=0; f := false;
  while (i<tab.ControlCount-1) or not f do begin
    f := tab.Controls[i].ClassType = TSynEdit;
    inc(i);
  end;
  if f then result := TSynEdit(tab.Controls[i-1]);
end;

function TXSDockManager.GetEditor: TSynEdit;
begin
  result := GetEditorByNum(fpagecontrol.ActivePage,0);
end;

procedure TXSDockManager.InvalidateParent(p: TControl);
begin
  while p.Parent<>NIL do p := p.Parent;
  p.Invalidate;
end;

procedure TXSDockManager.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Sender IS TXSScrollbox   then TXSScrollbox(Sender).SetFocus;
  if Sender IS TXSpagecontrol then TXSpagecontrol(Sender).SetFocus;
  if Sender IS TSynEdit       then TSynEdit(Sender).SetFocus;
  UpdateText;
  InvalidateParent(fpagecontrol);
end;

procedure TXSDockManager.PageChange(Sender: TObject);
var
  tab: TXSScrollbox;
begin
  if fpagecontrol.pagelist.count = 0 then exit;
  tab   := GetTabByNum(fpagecontrol.ActivePage,0);
  if tab = nil then exit;
  if tab.activetab then begin
    if Editor = nil then exit;
    if not Editor.Focused then begin
      Editor.SetFocus;
      UpdateText;
      InvalidateParent(Editor.Parent);
    end;
  end;
end;

procedure TXSDockManager.PageClose(Sender: TObject; idx: integer);
begin
  CloseFileByIdx(idx);
end;

procedure TXSDockManager.EditChange(Sender: TObject);
var
  tab   : TXSScrollBox;
begin
  tab := GetTabByNum(fpagecontrol.ActivePage,0);
  if tab       = nil then exit;
  if Editor    = nil then exit;
  tab.Modified := Editor.Modified;
  UpdateText;
  fpagecontrol.Invalidate;
end;

procedure TXSDockManager.ScrollChange(Sender: TObject);
var
  tab   : TXSScrollBox;
begin
  tab := GetTabByNum(fpagecontrol.ActivePage,0);
  if tab = nil then exit;
  Editor.TopLine  := tab.VScroll.Position+1;
  Editor.LeftChar := tab.HScroll.Position;
end;

procedure TXSDockManager.EditPaint(Sender: TObject; ACanvas: TCanvas);
var
  tab   : TXSScrollBox;
begin
  tab := GetTabByNum(fpagecontrol.ActivePage,0);
  if tab = nil then exit;
  tab.VScroll.Max      := Editor.Lines.Count-1;
  tab.VScroll.Position := Editor.TopLine-1;
  tab.VScroll.PageSize := Editor.Height div Editor.LineHeight-1;
  tab.HScroll.Max      := 1024;
  tab.HScroll.Position := Editor.LeftChar;
  tab.HScroll.PageSize := 80;
end;

function TXSDockManager.AllModified: boolean;
var
  tab  : TXSScrollBox;
  i,j  : integer;
begin
  result := false;
  i := 0;
  while (i<=fpagecontrol.pagelist.count-1) and not result do begin
    j := 0;
    while (j<=TStringList(fpagecontrol.pagelist.objects[i]).count-1) and not result do begin
      tab := GetTabByNum(i,j);
      if (tab <> NIL) then result := tab.Modified;
      inc(j);
    end;
    inc(i);
  end;
end;

procedure TXSDockManager.SwapCommand(idx1,idx2: TSynEditorMouseCommand);
var
  i: integer;
begin
  for i:= 0 to Editor.MouseActions.Count - 1  do
    if Editor.MouseActions.Items[i].Command = idx1 then
      Editor.MouseActions.Items[i].Command := idx2;
end;

procedure TXSDockManager.EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  i: integer;
begin
  if ssCtrl IN Shift then begin
    SwapCommand(emcNone,emcMouseLink);
    Editor.MouseOptions := [emUseMouseActions,emShowCtrlMouseLinks];
  end;
end;

procedure TXSDockManager.EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  i: integer;
begin
  if key=17 then begin
    SwapCommand(emcMouseLink,emcNone);
    Editor.MouseOptions := [];
  end;
end;

procedure TXSDockManager.EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
var
  i: integer;
begin
  SwapCommand(emcMouseLink,emcNone);
  Editor.MouseOptions := [];
  ShowMessage(wordatxy);
end;

procedure TXSDockManager.EditorMouseLink(Sender: TObject; X, Y: Integer;var AllowMouseLink: Boolean);
var
  Attri: TSynHighlighterAttributes;
  toktype,start: integer;
begin
  AllowMouseLink := true;
  Editor.GetHighlighterAttriAtRowColEx(Point(X,Y),wordatxy,toktype,start,Attri);
end;

procedure TXSDockManager.EditorOptions(ed: TSynEdit; tab: TXSScrollBox);
var
  SynMarkup: TSynEditMarkupHighlightAllCaret;
begin
  ed.Highlighter := TSynObrSyn.Create(ed);
  ed.PopupMenu   := Form1.PopupMenu1;
  ed.Options     := ed.Options + [eoDragDropEditing];
  //ed.Font.Name   := 'PT Mono';
  //ed.Font.Size   := 10;

  ed.Font.Quality:= fqNonAntialiased;
  ed.Width       := Form1.Width;
  ed.Height      := Form1.Height;
  ed.Parent      := tab;
  ed.Align       := alClient;
  ed.BorderStyle := bsNone;
  ed.ScrollBars  := ssNone;
  ed.LineHighlightColor.Background := $00DDFFFF;
  ed.SelectedColor.Background      := $00A56D53;
  ed.Gutter.Color                  := $E5E5E5;
  ed.BracketMatchColor.Foreground  := $0000FF;
  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).ShowOnlyLineNumbersMultiplesOf:=10;
  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).MarkupInfo.Foreground:=clGray; //Цыфры
  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).MarkupInfo.Background:=$E5E5E5;
  TSynGutterSeparator(ed.Gutter.Parts.Part[3]).MarkupInfo.Foreground:=ed.Color;  //Разделитель
  TSynGutterCodeFolding(ed.Gutter.Parts.Part[4]).MarkupInfo.Foreground:=clGray;
  TSynGutterCodeFolding(ed.Gutter.Parts.Part[4]).MarkupInfo.Background:=$E5E5E5;
  ed.FoldedCodeColor.FrameColor:=0;

  ed.OnMouseLink := @EditorMouseLink;
  ed.OnClickLink := @EditorClickLink;
  ed.OnMouseDown := @MouseDown;
  ed.OnPaint     := @EditPaint;
  ed.OnChange    := @EditChange;
  ed.OnKeyDown   := @EditorKeyDown;
  ed.OnKeyUp     := @EditorKeyUp;

  tab.HScroll.OnChange := @ScrollChange;
  tab.VScroll.OnChange := @ScrollChange;

  ed.MouseLinkColor.Style:=[fsUnderline];
  ed.MouseActions.Add;
  ed.MouseActions.Items[0].Shift        := [ssCtrl];
  ed.MouseActions.Items[0].ShiftMask    := [ssShift, ssAlt, ssCtrl];

  SynMarkup := TSynEditMarkupHighlightAllCaret(ed.MarkupByClass[TSynEditMarkupHighlightAllCaret]);
  SynMarkup.MarkupInfo.FrameColor := $E5E5E5;
  SynMarkup.MarkupInfo.Background := $E5E5E5;
  SynMarkup.WaitTime := 100;
  SynMarkup.Trim     := True;
  SynMarkup.FullWord := True;
  SynMarkup.IgnoreKeywords := False;

end;

procedure TXSDockManager.NewPage;
var
  i  : integer;
  s  : string;
  tab: TXSScrollbox;
  ed : TSynEdit;

  function SameCaption(AName: string): boolean;
  var
    i: integer;
  begin
    result := false;
    i:=0;
    while (i<fpagecontrol.pagelist.count) and not result do begin
      result := fpagecontrol.pagelist[i] = AName;
      inc(i);
    end;
  end;

begin
  i := 1;
  s := fdefname+'1'+fdefext;
  while (i<100) and SameCaption(s) do begin
    inc(i);
    s :=fdefname+IntToStr(i)+fdefext;
  end;

  fpagecontrol.AddPage(s,false);
  tab := fpagecontrol.AddTab('Code','');
  tab.OnMouseDown := @MouseDown;
  ed            := TSynEdit.Create(Form1);
  ed.Visible    := false;
  EditorOptions(ed,tab);
  ed.Text       := #10;
  ed.Visible    := true;
  PageChange(self);

  tab := fpagecontrol.AddTab('Design','');
  tab.OnMouseDown := @MouseDown;
end;

procedure TXSDockManager.OpenPage;
var
  ed    : TSynEdit;
  i,j   : integer;
  fn,fp : string;
  tab   : TXSScrollBox;
  sl    : TStringList;
begin
  fopendialog.Filter := ffilter;
  fopendialog.FilterIndex := 0;
  fopendialog.FileName:='';
  if fopendialog.Execute then begin

    for i:=0 to fopendialog.Files.count-1 do begin

      fn := ExtractFileName(fopendialog.Files[i]);
      fp := ExtractFilePath(fopendialog.Files[i]);

      tab := nil; j := 0;
      while (j<=fpagecontrol.pagelist.count-1) do begin
        tab := GetTabByNum(j,0);
        if (tab.Path = fp+fn) then break else tab := nil;
        inc(j);
      end;

      if tab <> NIL then begin
        fpagecontrol.ActivePage := j;
      end else begin
        fpagecontrol.AddPage(fn,true);
        tab := fpagecontrol.AddTab('Code',fp+fn);

        ed            := TSynEdit.Create(Form1);
        ed.Visible    := false;
        EditorOptions(ed,tab);

        sl := TStringList.Create;
        sl.LoadFromFile(fp+fn);
        ed.Text := sl.Text;
        sl.Free;

        ed.Visible    := true;
        PageChange(self);

        //tab := fpagecontrol.AddTab('Design','');
        //tab.OnMouseDown := @MouseDown;
      end;

    end;

  end;

end;

procedure TXSDockManager.ClosePage;
begin
  CloseFileByIdx(fpagecontrol.activepage);
  fpagecontrol.Invalidate;
end;

function TXSDockManager.SaveFileAsByIdx(idx: integer): Boolean;
var
  tab  : TXSScrollbox;
begin

  tab   := GetTabByNum(idx,0);
  if tab = nil then exit;

  fsavedialog.Filter := ffilter;
  fsavedialog.FilterIndex := 0;
  fsavedialog.FileName := fpagecontrol.pagelist[idx];
  Result := fsavedialog.Execute;

  if Result then begin

    if FileExists(fsavedialog.FileName) and (tab.Path<>fsavedialog.FileName) then
      case MessageDlg('A file "' + fsavedialog.FileName + '" already exists.'+#13+'Replace it?' , mtWarning, mbYesNoCancel, 0) of
        mrYes:      Result := True;
        mrNo:       Result := False;
        mrCancel:   Result := False;
      end;

    if Result and (tab.Path <> '') and (tab.Path<>fsavedialog.FileName) and FileExists(tab.Path) then
      case MessageDlg('Delete old file "' + tab.Path + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes: DeleteFile(tab.Path);
      end;

    if Result then begin
      if ExtractFileExt(fsavedialog.FileName) = '' then fsavedialog.FileName:=fsavedialog.FileName+fdefext;
      GetEditorByNum(idx,0).Lines.SaveToFile(fsavedialog.FileName);
      fpagecontrol.pagelist[idx] := ExtractFileName(fsavedialog.FileName);
      tab.Path    := fsavedialog.FileName;
      tab.Modified := false;
      //fpagecontrol.UpdatePage(idx);
      UpdateText;
      InvalidateParent(fpagecontrol.Parent);
    end;

  end;

end;

procedure TXSDockManager.SaveFileAs;
begin
  SaveFileAsByIdx(fpagecontrol.activepage);
end;

function TXSDockManager.SaveFileByIdx(idx: integer): Boolean;
var
  tab  : TXSScrollbox;
begin
  tab    := GetTabByNum(idx,0);
  Result := tab.Path <> '';
  if not Result then
    Result := SaveFileAsByIdx(idx)
  else begin
    GetEditorByNum(idx,0).Lines.SaveToFile(tab.path);
    tab.Path     := fsavedialog.FileName;
    tab.Modified := false;
    //fpagecontrol.UpdatePage(idx);
    UpdateText;
    InvalidateParent(fpagecontrol.parent);
  end;
end;

procedure TXSDockManager.SaveFile;
begin
  SaveFileByIdx(fpagecontrol.activepage);
end;

procedure TXSDockManager.SaveAll;
var
  i: integer;
begin
  for i:=pagecontrol.pagelist.count-1 downto 0 do
    if not SaveFileByIdx(i) then exit;
end;

function TXSDockManager.CloseFileByIdx(idx: integer): Boolean;
begin
  Result := not GetTabByNum(idx,0).Modified;
  if not Result then
    case MessageDlg('Save changed file "' + fpagecontrol.pagelist[idx] + '" ?' , mtWarning, mbYesNoCancel, 0) of
      mrYes:      Result := SaveFileByIdx(idx);
      mrNo:       Result := True;
      mrCancel:   Result := False;
    end;
  if result then fpagecontrol.ClosePage(idx);
end;

procedure TXSDockManager.CloseFile;
begin
  CloseFileByIdx(fpagecontrol.activepage);
end;

function TXSDockManager.CloseAll: Boolean;
begin
  Result := False;
  while fpagecontrol.pagelist.count > 0 do
    if not CloseFileByIdx(fpagecontrol.pagelist.count-1) then Exit;
  Result := True;
end;

function TXSDockManager.SaveCloseQuery: Boolean;
var
  i: integer;
begin
  Result:=False;
  for i:=fpagecontrol.pagelist.count-1 downto 0 do
    if GetTabByNum(i,0).Modified then
      case MessageDlg('Save changed file "' + fpagecontrol.pagelist[i] + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes   : if not SaveFileByIdx(i) then Exit;
        mrNo    : ;
        mrCancel: begin Result := False; Exit; end;
      end;
  Result:=True;
end;

procedure TXSDockManager.Cut;
begin
  Editor.CutToClipboard;
end;

procedure TXSDockManager.Copy;
begin
  Editor.CopyToClipboard;
end;

procedure TXSDockManager.Paste;
begin
  Editor.PasteFromClipboard;
end;

procedure TXSDockManager.Delete;
begin
  Editor.ClearSelection;
end;

procedure TXSDockManager.Undo;
begin
  Editor.Undo;
end;

procedure TXSDockManager.Redo;
begin
  Editor.Redo;
end;

procedure TXSDockManager.Find;
begin
  Form1.findpanel.Height:=40;
end;

procedure TXSDockManager.FindNext;
begin
end;

procedure TXSDockManager.FindPrev;
begin
end;

procedure TXSDockManager.Unindent;
begin
  Editor.CommandProcessor(ecBlockUnIndent,#0,nil);
end;

procedure TXSDockManager.Indent;
begin
  Editor.CommandProcessor(ecBlockIndent,#0,nil);
end;

procedure TXSDockManager.Build;
begin
end;

procedure TXSDockManager.Run;
begin
end;

procedure TXSDockManager.Pause;
begin
end;

procedure TXSDockManager.Stop;
begin
end;

procedure TXSDockManager.SelectAll;
begin
  Editor.SelectAll;
end;

procedure TXSDockManager.DeleteSelection;
begin
  Editor.CommandProcessor(ecLeft,#0,nil);
  Editor.CommandProcessor(ecRight,#0,nil);
end;

procedure TXSDockManager.Upper;
begin
  Editor.CommandProcessor(ecUpperCase,#0,nil);
end;

procedure TXSDockManager.Lower;
begin
  Editor.CommandProcessor(ecLowerCase,#0,nil);
end;

procedure TXSDockManager.Swap;
begin
  Editor.CommandProcessor(ecToggleCase,#0,nil);
end;

procedure TXSDockManager.Title;
begin
  Editor.CommandProcessor(ecTitleCase,#0,nil);
end;

end.

