unit Xscontrols;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, ExtCtrls, FileUtil, Forms, Math,
  LCLType, LCLintf, ImgList, Graphics, GraphType, Types, Dialogs;

type
  {$IFDEF CPU64}int = int64;  {$ENDIF}
  {$IFDEF CPU32}int = integer;{$ENDIF}
  TNotifyEvent = procedure(Sender: TObject) of object;

  { TXSToolButton }
  TXSToolBar = class;

  TXSToolButton = class(TObject)
  public
    Tag         : integer;
    ImageIndex  : integer;
    Parent      : TXSToolBar;
    constructor Create(TheOwner: TComponent);
  private
    X,Y         : integer;
    Width       : integer;
    FOnClick    : TNotifyEvent;
    FMode       : boolean;
    fEnabled    : boolean;
    procedure   SetEnabled(const AValue: boolean);
  published
    property    OnClick: TNotifyEvent read FOnClick write FOnClick;
    property    Mode: boolean read FMode write FMode;
    property    Enabled: boolean read fEnabled write SetEnabled;
  end;

  { TXSToolBar }

  TXSToolBar = class(TCustomControl)
  private
    FImages     : TCustomImageList;
    FList       : TStringList;
    fCount      : integer;
    foffset     : integer;
    fdown       : boolean;
    tmpoffs     : integer;
    procedure   SetImages(const AValue: TCustomImageList);
    procedure   GetButtonAtMousePos(X,Y: Integer);
  protected
    procedure   Paint; override;
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure   OnShowHintSelf(Sender: TObject; HintInfo: PHintInfo);
    procedure   MouseEnter; override;
    procedure   MouseLeave; override;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   AddButton(HintName: string; tb: TXSToolButton);
    procedure   AddSeparator;
  published
    property    Images: TCustomImageList read FImages write SetImages;
    property    Count : integer read fCount write fCount;
  end;

  { TXSScrollbar }

  TArrowType = (
    atArrowUp,
    atArrowDown,
    atArrowLeft,
    atArrowRight
    );

  TXSScrollbar = class(TCustomControl)
  private
    FKind       : TScrollBarKind;
    FPos        : Integer;
    FMin        : Integer;
    FMax        : Integer;
    FPage       : Integer;

    FIn         : TRect;
    FInUp       : TRect;
    FInDown     : TRect;
    FInThumb    : TRect;
    FInCorner   : TRect;
    FInPageUp   : TRect;
    FInPageDown : TRect;

    FMouseDown,
    FMouseDownOnUp,
    FMouseDownOnDown,
    FMouseDownOnThumb,
    FMouseDownOnPageUp,
    FMouseDownOnPageDown: boolean;

    FMouseDragOffset: Integer;
    FTimerDelay : Integer;
    FTimer      : TTimer;
    FIndentCorner: Integer;

    FOnChange   : TNotifyEvent;
    procedure   SetKind(Value: TScrollBarKind);
    procedure   SetPos(Value: Integer);
    procedure   SetMin(Value: Integer);
    procedure   SetMax(Value: Integer);
    procedure   SetPage(Value: Integer);
    procedure   DoScrollBy(NDelta: Integer);
    procedure   TimerTimer(Sender: TObject);
    function    MouseToPos(X, Y: Integer): Integer;
    procedure   DoUpdatePosOnDrag(X, Y: Integer);
    procedure   PaintArrow(C: TCanvas; R: TRect; Typ: TArrowType);
    procedure   PaintThumb(C: TCanvas);
    function    GetPxAtScroll(APos: Integer): Integer;
    procedure   UpdateThumbRect;
    procedure   UpdateCornerRect;
    procedure   PaintMark(C: TCanvas; P: TPoint; Clr: integer);
  public
    constructor Create(AOnwer: TComponent); override;
    destructor  Destroy; override;
    property    Position: Integer read FPos write SetPos;
    property    Min: Integer read FMin write SetMin;
    property    Max: Integer read FMax write SetMax;
    property    PageSize: Integer read FPage write SetPage;
  protected
    procedure   Paint; override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
  published
    property    Kind: TScrollBarKind read FKind write SetKind;
    property    TimerDelay: Integer read FTimerDelay write FTimerDelay;
    property    IndentCorner: Integer read FIndentCorner write FIndentCorner;
    property    OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  {TXScrollbox}

  { TXSScrollbox }

  TXSScrollbox = class(TCustomControl)
  private
    fpath      : string;
    fcaption   : string;
    fmodified  : boolean;
    factivetab : boolean;
    fonmousedown: TMouseEvent;
  protected
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
  public
    VScroll     : TXSScrollbar;
    HScroll     : TXSScrollbar;
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property    Path    : string read fpath write fpath;
    property    Caption : string read fcaption write fcaption;
    property    Modified: boolean read fmodified write fmodified;
    property    OnMouseDown: TMouseEvent read fonmousedown write fonmousedown;
    property    ActiveTab: boolean read factivetab write factivetab;
  end;

  { TXSPagecontrol }

  TPageCloseEvent = procedure (Sender: TObject; idx: Integer) of object;

  TXSPagecontrol = class(TCustomControl)
  private
    ftext       : TStringList;
    fcanvasrect : TRect;
    fcradius    : integer;
    frectoffset : integer;
    fpagewidth  : integer;
    fpageheight : integer;
    fpageoffset : integer;
    ftabwidth   : integer;
    ftabheight  : integer;
    ftaboffset  : integer;
    fcurtabidx  : integer;
    factivepage : integer;
    fmaxvisible : integer;
    fmaxtab     : integer;
    flrrect     : TRect;
    flrwidth    : integer;
    flrvisible  : boolean;
    flbactive   : boolean;
    flractive   : boolean;
    fstartpage  : integer;
    fendpage    : integer;

    fbgcolor    : integer;
    ffgcolor    : integer;
    ffrcolor    : integer;
    fabgcolor   : integer;
    fafgcolor   : integer;
    fafrcolor   : integer;

    fpageangle  : integer;
    fclosewidth : integer;
    fmousedown  : boolean;
    fmousemove  : boolean;
    fclosind    : boolean;
    fcloseidx   : integer;
    fonchange   : TNotifyEvent;
    fonmousedown: TMouseEvent;
    fonpageclose: TPageCloseEvent;
    procedure   SetPage(Value: Integer);
  protected
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure   Paint; override;
    procedure   InvisibleTab(idx: integer);
    procedure   InvisibleActiveTab;
    procedure   DrawPage(acaption: string; r: TRect;  flag,closed,modified: boolean);
    procedure   DrawTab(AText: string; r: TRect; flag: boolean);
    procedure   DrawText;
  public
    pagelist    : TStringList;
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   ClosePage(idx: integer);
    procedure   AddText(AText: string; AWidth: integer);
    procedure   AddPage(ACaption: string; refreshing: boolean);
    function    AddTab(ACaption, APath: string): TXSScrollbox;
  published
    property    Text: TStringList read ftext write ftext;
    property    ActivePage : integer read factivepage write SetPage;
    property    OnChange   : TNotifyEvent read fonchange write fonchange;
    property    OnMouseDown: TMouseEvent read fonmousedown write fonmousedown;
    property    OnPageClose: TPageCloseEvent read fonpageclose write fonpageclose;
  end;

  {TXSPanel}

  TXSPanel= class(TCustomControl)
  private
    fclose      : boolean;
    fdown       : boolean;
    closerect   : TRect;
    FImages     : TCustomImageList;
    hdrheight   : integer;
    procedure   SetImages(const AValue: TCustomImageList);
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
  public
    ImageIndex  : integer;
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property    Images: TCustomImageList read FImages write SetImages;
  end;

implementation

{ TXSPanel }

constructor TXSPanel.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  ImageIndex := -1;
  hdrheight  := 22;
  width      := 100;
  height     := 100;
end;

destructor TXSPanel.Destroy;
begin

  inherited Destroy;
end;

procedure TXSPanel.SetImages(const AValue: TCustomImageList);
begin
  if FImages<>AValue then
    FImages:=AValue;
end;

procedure TXSPanel.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  p: TControl;
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown := true;
  SetFocus;
  p := Parent;
  while p.Parent<>NIL do p := p.Parent;
  p.Invalidate;
end;

procedure TXSPanel.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if PtInRect(closerect,Point(X,Y)) then visible := false;
  fdown := false;
  fclose := false;
  Invalidate;
end;

procedure TXSPanel.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if fdown then exit;
  fclose := PtInRect(closerect,Point(X,Y));
  Invalidate;
end;

procedure TXSPanel.Resize;
begin
  Invalidate;
end;

procedure TXSPanel.Paint;
var
  i: integer;
begin
  inherited Paint;

  Canvas.Pen.Color:=$E0E0E0;
  if Focused then
    Canvas.Brush.Color:= $E0E0E0
  else
    Canvas.Brush.Color:= $E5E5E5;

  Canvas.Font.Color := clGray;
  Canvas.RoundRect(2,2,width-2,hdrheight+2,5,5);
  i := 10;
  if (FImages<>NIL) and (ImageIndex>-1) then begin
    Images.Draw(Canvas,5,(hdrheight - FImages.Height) div 2 + 2,ImageIndex,Focused);
    inc(i,FImages.Width);
  end;
  Canvas.TextOut(i,(hdrheight - FImages.Height) div 2 + 2,Caption);

  closerect := Rect(width-2-5-10,(hdrheight - 10) div 2 + 2,width-2-5,(hdrheight - 10) div 2 + 12);

  if fclose then
    Canvas.Pen.Color  := clRed
  else
    Canvas.Pen.Color  := clGray;

  Canvas.Line(closerect.Left+1,closerect.Top+1,closerect.Right,closerect.Bottom);
  Canvas.Line(closerect.Left+1,closerect.Top+2,closerect.Right-1,closerect.Bottom);
  Canvas.Line(closerect.Left+2,closerect.Top+1,closerect.Right,closerect.Bottom-1);

  Canvas.Line(closerect.Left+1,closerect.Bottom-1,closerect.Right,closerect.Top);
  Canvas.Line(closerect.Left+1,closerect.Bottom-2,closerect.Right-1,closerect.Top);
  Canvas.Line(closerect.Left+2,closerect.Bottom-1,closerect.Right,closerect.Top+1);

end;

{TXSToolButton}

constructor TXSToolButton.Create(TheOwner: TComponent);
begin
  inherited Create;
  Parent := TXSToolBar(TheOwner);
  ImageIndex := -1;
  X   := 0;
  Y   := 0;
  tag := 0;
  width := 25;
  Enabled := true;
  FMode := false;
end;

procedure TXSToolButton.SetEnabled(const AValue: boolean);
begin
  if AValue<>fEnabled then begin
    fEnabled := AValue;
    if Parent <> nil then
      InvalidateRect(Parent.handle,NIL,false);
  end;
end;

constructor TXSToolBar.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  Height := 27;
  Align  := alTop;
  Images := TCustomImageList.Create(Self);
  FList  := TStringList.Create;
  fCount := -1;
  foffset:= -1;
  DoubleBuffered:=true;
  fdown := false;
  Color := clWindow;
  ShowHint := true;
  OnShowHint := @OnShowHintSelf;
end;

destructor TXSToolBar.Destroy;
var
  i: integer;
begin
  for i:=0 to FList.Count-1 do FList.Objects[i].Free;
  FreeAndNil(FList);
  inherited Destroy;
end;

procedure TXSToolBar.SetImages(const AValue: TCustomImageList);
begin
  if FImages<>AValue then
    FImages:=AValue;
end;

procedure TXSToolBar.AddButton(HintName: string; tb: TXSToolButton);
begin
  tb.X := 2;
  if FList.Count>0 then tb.X := TXSToolButton(FList.Objects[FList.Count-1]).X+TXSToolButton(FList.Objects[FList.Count-1]).width+2;
  tb.Y := 2;
  tb.Mode := false;
  FList.AddObject(HintName,tb);
end;

procedure TXSToolBar.AddSeparator;
var
  tb: TXSToolButton;
begin
  tb := TXSToolButton.Create(Self);
  tb.Width:=3;
  tb.X := 2;
  if FList.Count>0 then tb.X := TXSToolButton(FList.Objects[FList.Count-1]).X+TXSToolButton(FList.Objects[FList.Count-1]).width+2;
  tb.Y := 4;
  tb.Mode := true;
  FList.AddObject('',tb);
end;

procedure TXSToolBar.GetButtonAtMousePos(X,Y: Integer);
var
  i: integer;
  flag: boolean;
begin
  i:=0;
  flag := false;
  while (i<FList.Count) and not flag do begin
    if (X>=TXSToolButton(FList.Objects[i]).X) and (X<=TXSToolButton(FList.Objects[i]).X+TXSToolButton(FList.Objects[i]).width)
    and (Y>=TXSToolButton(FList.Objects[i]).Y) and (Y<=TXSToolButton(FList.Objects[i]).Y+height-TXSToolButton(FList.Objects[i]).Y)
    and not TXSToolButton(FList.Objects[i]).mode then begin
      foffset := i;
      flag := true;
    end;
    inc(i);
  end;
end;

procedure TXSToolBar.OnShowHintSelf(Sender: TObject; HintInfo: PHintInfo);
var
  Pos : TPoint;
begin
  if foffset=-1 then exit;
  with HintInfo^ do begin
    Pos.X := TXSToolButton(FList.Objects[foffset]).X + TXSToolButton(FList.Objects[foffset]).width;
    Pos.Y := -TXSToolButton(FList.Objects[foffset]).width;
    HintPos := ClientToScreen(Pos);
    HintStr := FList[foffset];
  end;
end;

procedure TXSToolBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  GetButtonAtMousePos(X,Y);
  tmpoffs:=foffset;
  fdown := Button=mbLeft;
  InvalidateRect(handle,NIL,false);
end;

procedure TXSToolBar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  foffset := -1;
  if fdown then exit;
  GetButtonAtMousePos(X,Y);
  if foffset>-1 then begin
    Application.CancelHint;
    if TXSToolButton(FList.Objects[foffset]).Enabled then
      Application.ActivateHint(ClientToScreen(Point(0,0)));
  end;
  InvalidateRect(handle,NIL,false);
end;

procedure TXSToolBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  GetButtonAtMousePos(X,Y);
  if fdown and (foffset=tmpoffs) and (foffset>-1) then begin
    fdown := false;
    InvalidateRect(handle,NIL,false);
    if TXSToolButton(FList.Objects[foffset]).Enabled and Assigned(TXSToolButton(FList.Objects[foffset]).FOnClick) then TXSToolButton(FList.Objects[foffset]).FOnClick(TXSToolButton(FList.Objects[foffset]));
  end;
  fdown := false;
  InvalidateRect(handle,NIL,false);
end;

procedure TXSToolBar.MouseEnter;
begin
  inherited MouseEnter;
  foffset := -1;
  InvalidateRect(handle,NIL,false);
end;

procedure TXSToolBar.MouseLeave;
begin
  inherited MouseLeave;
  foffset := -1;
  InvalidateRect(handle,NIL,false);
end;

procedure TXSToolBar.Paint;
var
  i,j: integer;
begin
  for i:=0 to FList.Count-1 do begin
    if TXSToolButton(FList.Objects[i]).Enabled then begin
      Canvas.Pen.Color  := $E0E0E0;
      Canvas.Brush.Color:= $E0E0E0;
    end else begin
      Canvas.Pen.Color  := clWhite;
      Canvas.Brush.Color:= clWhite;
    end;
    j:=0;
    if TXSToolButton(FList.Objects[i]).Enabled and (foffset = i) then begin
      if fdown then begin
        Canvas.Pen.Color  := $D8D8D8;
        Canvas.Brush.Color:= $D8D8D8;
        j:=1;
      end else begin
        Canvas.Pen.Color  := $D0D0D0;
        Canvas.Brush.Color:= $D0D0D0;
      end;
    end;
    Canvas.RoundRect(TXSToolButton(FList.Objects[i]).X,TXSToolButton(FList.Objects[i]).Y,TXSToolButton(FList.Objects[i]).X+TXSToolButton(FList.Objects[i]).width,height-TXSToolButton(FList.Objects[i]).Y,4,3);
    if (Images<>NIL) and (TXSToolButton(FList.Objects[i]).ImageIndex>-1) then
      Images.Draw(Canvas,TXSToolButton(FList.Objects[i]).X+(TXSToolButton(FList.Objects[i]).width-FImages.Width) div 2,(height - FImages.height) div 2 + j,TXSToolButton(FList.Objects[i]).ImageIndex,TXSToolButton(FList.Objects[i]).enabled);
  end;
end;

{ TXSScrollbar }

constructor TXSScrollbar.Create(AOnwer: TComponent);
begin
  inherited Create(AOnwer);
  Width  := 200;
  Height := 15;
  FKind  := sbHorizontal;
  FMin   := 0;
  FMax   := 100;
  FPage  := 20;
  Color  := $E8E8E8;
  FMouseDown  := false;
  FMouseDragOffset:= 0;
  FTimerDelay:= 80;
  FTimer:= TTimer.Create(Self);
  FTimer.Enabled:= false;
  FTimer.Interval:= FTimerDelay;
  FTimer.OnTimer:= @TimerTimer;
  FIndentCorner:= 0;
end;

destructor TXSScrollbar.Destroy;
begin
  FTimer.Enabled:= false;
  inherited Destroy;
end;

procedure TXSScrollbar.SetKind(Value: TScrollBarKind);
begin
  if Value<>FKind then begin
    FKind:= Value;
    Invalidate;
  end;
end;

procedure TXSScrollbar.SetPos(Value: Integer);
begin
  Value:= Math.Min(Value, FMax);
  Value:= Math.Max(Value, FMin);
  if FPos<>Value then begin
    FPos:= Value;
    Invalidate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TXSScrollbar.SetMin(Value: Integer);
begin
  if FMin<>Value then begin
    FMin:= Value;
    FPos:= Math.Max(FPos, FMin);
    Invalidate;
  end;
end;

procedure TXSScrollbar.SetMax(Value: Integer);
begin
  if FMax<>Value then begin
    FMax:= Value;
    FPos:= Math.Min(FPos, FMax);
    Invalidate;
  end;
end;

procedure TXSScrollbar.SetPage(Value: Integer);
begin
  if FPage<>Value then begin
    FPage:= Value;
    Invalidate;
  end;
end;

function TXSScrollbar.MouseToPos(X, Y: Integer): Integer;
begin
  if Kind=sbHorizontal then
    Result:= FMin + (X-FIn.Left) * (FMax-FMin) div Math.Max(FIn.Right-FIn.Left, 1)
  else
    Result:= FMin + (Y-FIn.Top) * (FMax-FMin) div Math.Max(FIn.Bottom-FIn.Top, 1);
end;

function TXSScrollbar.GetPxAtScroll(APos: Integer): Integer;
var
  N0, NLen: Integer;
begin
  if Kind=sbHorizontal then begin
    N0:= FIn.Left;
    NLen:= FIn.Right-FIn.Left
  end else begin
    N0:= FIn.Top;
    NLen:= FIn.Bottom-FIn.Top;
  end;
  Result:= N0 + (APos-FMin) * NLen div Math.Max(1, FMax-FMin);
end;

procedure TXSScrollbar.DoUpdatePosOnDrag(X, Y: Integer);
var
  N: Integer;
begin
  N:= MouseToPos(
    X-FMouseDragOffset,
    Y-FMouseDragOffset);
  N:= Math.Max(N, FMin);
  N:= Math.Min(N, FMax-FPage);
  SetPos(N);
end;

procedure TXSScrollbar.DoScrollBy(NDelta: Integer);
var
  N: Integer;
begin
  N:= FPos;
  Inc(N, NDelta);
  if (NDelta>0) then
    N:= Math.Min(N, FMax-FPage);
  SetPos(N);
end;

procedure TXSScrollbar.TimerTimer(Sender: TObject);
var
  P: TPoint;
begin
  P:= Mouse.CursorPos;
  P:= ScreenToClient(P);

  if FMouseDownOnDown and PtInRect(FInDown, P) then
    DoScrollBy(1)
  else
  if FMouseDownOnUp and PtInRect(FInUp, P) then
    DoScrollBy(-1)
  else
  if FMouseDownOnPageDown and PtInRect(FInPageDown, P) then
    DoScrollBy(FPage)
  else
  if FMouseDownOnPageUp and PtInRect(FInPageUp, P) then
    DoScrollBy(-FPage);
end;

procedure TXSScrollbar.MouseDown(Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
begin
  FMouseDown:= Button=mbLeft;
  FMouseDownOnThumb:= PtInRect(FInThumb, Point(X, Y));
  FMouseDownOnUp:= PtInRect(FInUp, Point(X, Y));
  FMouseDownOnDown:= PtInRect(FInDown, Point(X, Y));
  FMouseDownOnPageUp:= PtInRect(FInPageUp, Point(X, Y));
  FMouseDownOnPageDown:= PtInRect(FInPageDown, Point(X, Y));

  if Kind=sbHorizontal then
    FMouseDragOffset:= X-FInThumb.Left
  else
    FMouseDragOffset:= Y-FInThumb.Top;

  FTimer.Enabled:= FMouseDown and
    (FMouseDownOnUp or
     FMouseDownOnDown or
     FMouseDownOnPageUp or
     FMouseDownOnPageDown);
end;

procedure TXSScrollbar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if FMouseDownOnThumb then begin
    DoUpdatePosOnDrag(X, Y);
    Exit;
  end;
end;

procedure TXSScrollbar.MouseUp(Button: TMouseButton; Shift: TShiftState;X, Y: Integer);
begin
  FMouseDown:= false;
  FMouseDownOnThumb:= false;
  FTimer.Enabled:= false;
  Invalidate;
end;

procedure TXSScrollbar.PaintArrow(C: TCanvas; R: TRect; Typ: TArrowType);
var
  P, P1, P2, P3: TPoint;
  cc: Integer;
begin
  C.Brush.Color:= $E8E8E8;
  C.FillRect(R);

  P:= CenterPoint(R);
  cc:= 3;

  case Typ of
    atArrowUp:
      begin
        P1:= Point(P.X-cc, P.Y+cc div 2);
        P2:= Point(P.X+cc, P.Y+cc div 2);
        P3:= Point(P.X, P.Y-cc+cc div 2);
      end;
    atArrowDown:
      begin
        P1:= Point(P.X-cc, P.Y-cc div 2);
        P2:= Point(P.X+cc, P.Y-cc div 2);
        P3:= Point(P.X, P.Y+cc-cc div 2);
      end;
    atArrowLeft:
      begin
        P1:= Point(P.X+cc div 2, P.Y-cc);
        P2:= Point(P.X+cc div 2, P.Y+cc);
        P3:= Point(P.X-cc+cc div 2, P.Y);
      end;
    atArrowRight:
      begin
        P1:= Point(P.X-cc div 2    -1, P.Y-cc);
        P2:= Point(P.X-cc div 2    -1, P.Y+cc);
        P3:= Point(P.X+cc-cc div 2 -1, P.Y);
      end;
    else
      Exit;
 end;

  C.Brush.Color:= clSilver;
  C.Pen.Color:= clSilver;
  C.Polygon([P1, P2, P3]);
end;

procedure TXSScrollbar.UpdateCornerRect;
begin
  FInCorner:= Rect(0, 0, 0, 0);
  if Kind=sbHorizontal then begin
    if FIndentCorner>0 then begin
      FInCorner:= Rect(ClientWidth-FIndentCorner, 0, ClientWidth, ClientHeight);
      Dec(FIn.Right, FIndentCorner);
    end else if FIndentCorner<0 then begin
      FInCorner:= Rect(0, 0, Abs(FIndentCorner), ClientHeight);
      Inc(FIn.Left, Abs(FIndentCorner));
    end;
  end else begin
    if FIndentCorner>0 then begin
      FInCorner:= Rect(0, ClientHeight-FIndentCorner, ClientWidth, ClientHeight);
      Dec(FIn.Bottom, FIndentCorner);
    end else if FIndentCorner<0 then begin
      FInCorner:= Rect(0, 0, ClientWidth, Abs(FIndentCorner));
      Inc(FIn.Top, Abs(FIndentCorner));
    end;
  end;
end;

procedure TXSScrollbar.UpdateThumbRect;
const
  cMinView = 15;
var
  R: TRect;
begin
  FInThumb:= Rect(0, 0, 0, 0);
  FInPageUp:= Rect(0, 0, 0, 0);
  FInPageDown:= Rect(0, 0, 0, 0);

  if Kind=sbHorizontal then begin
    if FIn.Right-FIn.Left<cMinView then Exit;
    R.Top:= FIn.Top;
    R.Bottom:= FIn.Bottom;
    R.Left:= GetPxAtScroll(FPos);
    R.Left:= Math.Min(R.Left, FIn.Right-cMinView);
    R.Right:= GetPxAtScroll(FPos+FPage);
    R.Right:= Math.Max(R.Right, R.Left+cMinView);
    R.Right:= Math.Min(R.Right, FIn.Right);
  end else begin
    if FIn.Bottom-FIn.Top<cMinView then Exit;
    R.Left:= FIn.Left;
    R.Right:= FIn.Right;
    R.Top:= GetPxAtScroll(FPos);
    R.Top:= Math.Min(R.Top, FIn.Bottom-cMinView);
    R.Bottom:= GetPxAtScroll(FPos+FPage);
    R.Bottom:= Math.Max(R.Bottom, R.Top+cMinView);
    R.Bottom:= Math.Min(R.Bottom, FIn.Bottom);
  end;
  FInThumb:= R;

  if Kind=sbHorizontal then begin
    FInPageUp:= Rect(FIn.Left, FIn.Top, FInThumb.Left, FIn.Bottom);
    FInPageDown:= Rect(FInThumb.Right, FIn.Top, FIn.Right, FIn.Bottom);
  end else begin
    FInPageUp:= Rect(FIn.Left, FIn.Top, FIn.Right, FInThumb.Top);
    FInPageDown:= Rect(FIn.Left, FInThumb.Bottom, FIn.Right, FIn.Bottom);
  end;
end;

procedure TXSScrollbar.PaintMark(C: TCanvas; P: TPoint; Clr: integer);
var
  i,j: integer;
begin
  for i:=0 to 4 do
    for j:=0 to 4 do
      C.Pixels[P.X-4+i*2,P.Y-4+j*2]:=Clr;
end;

procedure TXSScrollbar.PaintThumb(C: TCanvas);
const
  cMinMark = 15;
begin
  if fPage>=fMax then exit;
  C.Brush.Color:= $E0E0E0;
  C.Pen.Color:= clSilver;
  C.RoundRect(FInThumb,10,10);
  if Kind=sbHorizontal then begin
    if (FInThumb.Right-FInThumb.Left)>=cMinMark then
      PaintMark(C,CenterPoint(FInThumb),clSilver);
  end else begin
    if (FInThumb.Bottom-FInThumb.Top)>=cMinMark then
      PaintMark(C,CenterPoint(FInThumb),clSilver);
  end;
end;

procedure TXSScrollbar.Paint;
var
  fSize: Integer;
begin
  FIn:= ClientRect;

  UpdateCornerRect;
  if not IsRectEmpty(FInCorner) then begin
    PaintMark(Canvas,CenterPoint(FInCorner),clSilver);
  end;

  if Kind=sbHorizontal then begin
    fSize:= Math.Min(FIn.Bottom-FIn.Top, (FIn.Right-FIn.Left) div 2);
    FInUp:= Rect(FIn.Left, FIn.Top, FIn.Left+FSize, FIn.Bottom);
    FInDown:= Rect(FIn.Right-FSize, FIn.Top, FIn.Right, FIn.Bottom);
    PaintArrow(Canvas, FInUp, atArrowLeft);
    PaintArrow(Canvas, FInDown, atArrowRight);
    Inc(FIn.Left, fSize);
    Dec(FIn.Right, fSize);
  end else begin
    FSize:= Math.Min(FIn.Right-FIn.Left, (FIn.Bottom-FIn.Top) div 2);
    FInUp:= Rect(FIn.Left, FIn.Top, FIn.Right, FIn.Top+FSize);
    FInDown:= Rect(FIn.Left, FIn.Bottom-FSize, FIn.Right, FIn.Bottom);
    PaintArrow(Canvas, FInUp, atArrowUp);
    PaintArrow(Canvas, FInDown, atArrowDown);
    Inc(FIn.Top, fSize);
    Dec(FIn.Bottom, fSize);
  end;

  UpdateThumbRect;
  PaintThumb(Canvas);
end;

{ TXSScrollbox }

constructor TXSScrollbox.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  VScroll := TXSScrollbar.Create(Self);
  VScroll.Parent := Self;
  VScroll.Kind   := sbVertical;
  VScroll.Width  := 15;
  VScroll.Align  := alRight;

  HScroll        := TXSScrollbar.Create(Self);
  HScroll.Parent := self;
  HScroll.Kind   := sbHorizontal;
  HScroll.Align  := alBottom;
  HScroll.IndentCorner := VScroll.Width;
end;

destructor TXSScrollbox.Destroy;
begin
  FreeAndNil(VScroll);
  FreeAndNil(HScroll);
  inherited Destroy;
end;

procedure TXSScrollbox.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Assigned(fonmousedown) then fonmousedown(Self,Button,Shift,X,Y);
end;

{ TXSPagecontrol }

constructor TXSPagecontrol.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  ftext      := TStringList.Create;
  pagelist   := TStringList.Create;
  frectoffset := 2;
  fpageheight := 0;
  fcradius    := 5;
  fpageoffset := fcradius*2;
  fpagewidth  := 100;
  fpageheight := 27;
  ftabwidth   := 50;
  ftabheight  := 26;
  ftaboffset  := fcradius*2;
  fmaxtab     := 3;
  fcanvasrect := Rect(0,0,0,0);
  flrrect     := Rect(0,0,0,0);
  factivepage := -1;
  fmaxvisible := 0;
  flrwidth    := 36;
  flrvisible  := false;
  fstartpage  := 0;
  fendpage    := 0;

  fbgcolor    := $E5E5E5;
  ffgcolor    := clGray;
  ffrcolor    := clSilver;
  fabgcolor   := clWhite;
  fafgcolor   := clDefault;
  fafrcolor   := clSilver;

  fpageangle  := 3;
  fclosewidth := 10;
end;

destructor TXSPagecontrol.Destroy;
var
  i: integer;
begin
  for i:=0 to pagelist.count-1 do pagelist.objects[i].Free;
  FreeAndNil(pagelist);
  FreeAndNil(ftext);
  inherited Destroy;
end;

procedure TXSPagecontrol.SetPage(Value: Integer);
begin
  if factivepage<>Value then begin
    factivepage:=Value;
    InvisibleActiveTab;
    Invalidate;
  end;
end;

procedure TXSPagecontrol.ClosePage(idx: integer);
var
  i: integer;
  page: TStringList;
begin
  for i:=0 to TStringList(pagelist.objects[idx]).count-1 do
    TStringList(pagelist.objects[idx]).objects[i].free;
  pagelist.Objects[idx].Free;
  pagelist.Delete(idx);
  if factivepage>idx then dec(factivepage);
  if factivepage>pagelist.count-1 then factivepage:=pagelist.count-1;
  InvisibleActiveTab;
end;

procedure TXSPagecontrol.InvisibleTab(idx: integer);
var
  i,j : integer;
  page: TStringList;
begin
  for i:=0 to pagelist.count-1 do begin
    page := TStringList(pagelist.objects[i]);
    for j:=0 to page.count - 1 do begin
      TXSScrollbox(page.objects[j]).visible := false;
    end;
  end;
  page := TStringList(pagelist.objects[factivepage]);
  for j:=0 to page.count - 1 do
    if (j=idx) then begin
      TXSScrollbox(page.objects[j]).visible := true;
      TXSScrollbox(page.objects[j]).activetab := true;
    end else
      TXSScrollbox(page.objects[j]).activetab := false;
end;

procedure TXSPagecontrol.InvisibleActiveTab;
var
  i: integer;
  page: TStringList;
begin
  if factivepage>=0 then begin
    page := TStringList(pagelist.objects[factivepage]);
    i := 0;
    while (i <= page.count-1) and not TXSScrollbox(page.objects[i]).activetab do inc(i);
    InvisibleTab(i);
  end;
end;

procedure TXSPagecontrol.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  pagezone : TRect;
  pageidx  : integer;
  i        : integer;
  page     : TStringList;
begin
  inherited MouseDown(Button, Shift, X, Y);
  pagezone  := Rect(fpageoffset,0,fpageoffset+(fendpage-fstartpage)*fpagewidth,fpageheight);
  fmousedown := (Button=mbLeft) and PtInRect(pagezone,Point(X,Y));
  pageidx := (X-fpageoffset) div fpagewidth + fstartpage;
  if fmousedown and not fclosind then begin
    factivepage := pageidx;
    InvisibleActiveTab;
  end;
  if Assigned(fonmousedown) then fonmousedown(Self,Button,Shift,X,Y);
  Invalidate;
end;

procedure TXSPagecontrol.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  pagezone : TRect;
  closezone: TRect;
  tabzone  : TRect;
  pageidx  : integer;
  r        : TRect;
begin
  inherited MouseMove(Shift, X, Y);

  pagezone  := Rect(fpageoffset,0,fpageoffset+(fendpage-fstartpage)*fpagewidth,fpageheight);
  pageidx   := (X-fpageoffset) div fpagewidth + fstartpage;

  if not fmousedown and not fmousemove then begin
    closezone := Rect(fpageoffset+(pageidx-fstartpage)*fpagewidth+fpagewidth-fpageangle*2-fclosewidth,(fpageheight-fclosewidth) div 2+1,fpageoffset+(pageidx-fstartpage)*fpagewidth+fpagewidth-fpageangle*2,fpageheight-(fpageheight-fclosewidth) div 2 + 1);
    fclosind  := PtInRect(closezone,Point(X,Y));
    fcloseidx := -1;
    if fclosind then fcloseidx := pageidx;
  end;

  if fmousedown and not fclosind then begin
    fmousemove := true;
    if X<pagezone.left then dec(fstartpage);
    if X>pagezone.right then inc(fstartpage);
    if (pageidx<>factivepage) and (pageidx<pagelist.count) and (pageidx>=0) then begin
      pagelist.Exchange(factivepage,pageidx);
      factivepage := pageidx;
    end;
  end;

  fcurtabidx := -1;
  if factivepage >=0 then begin
    tabzone := Rect(ftaboffset-1,height-ftabheight,ftaboffset+TStringList(pagelist.Objects[factivepage]).count*ftabwidth,height);
    if PtInRect(tabzone,Point(X,Y)) then fcurtabidx := (X-ftaboffset) div ftabwidth;
  end;

  flbactive  := false;
  flractive  := false;
  if flrvisible and PtInRect(flrrect,Point(X,Y)) then begin
    r := flrrect;
    r.Right:=flrrect.Left+(flrrect.Right-flrrect.Left) div 2;
    flbactive  := PtInRect(r,Point(X,Y));
    flractive  := not flbactive;
  end;

  Invalidate;
end;

procedure TXSPagecontrol.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  closezone: TRect;
  pageidx  : integer;
begin
  inherited MouseUp(Button, Shift, X, Y);

  if fclosind and fmousedown and (fcloseidx <> -1) then begin
    pageidx   := (X-fpageoffset) div fpagewidth + fstartpage;
    closezone := Rect(fpageoffset+(pageidx-fstartpage)*fpagewidth+fpagewidth-fpageangle*2-fclosewidth,(fpageheight-fclosewidth) div 2+1,fpageoffset+(pageidx-fstartpage)*fpagewidth+fpagewidth-fpageangle*2,fpageheight-(fpageheight-fclosewidth) div 2 + 1);
    if PtInRect(closezone,Point(X,Y)) then begin
      if Assigned(fonpageclose) then
        fonpageclose(Self, fcloseidx)
      else
        ClosePage(fcloseidx);
      fclosind  := PtInRect(closezone,Point(X,Y));
      if fclosind then
        if fstartpage>0 then
          fcloseidx := pageidx-1
        else
          fcloseidx := pageidx;
      fmousedown := false;
      fmousemove := false;
      Invalidate;
      Exit;
    end;
  end;

  if flbactive then dec(fstartpage);
  if flractive then inc(fstartpage);

  if fcurtabidx > -1 then InvisibleTab(fcurtabidx);

  fmousedown := false;
  fclosind   := false;
  fmousemove := false;
  fcloseidx  := -1;
  fcurtabidx := -1;
  if Assigned(fonchange) then fonchange(self);
  Invalidate;
end;

procedure TXSPagecontrol.AddText(AText: string; AWidth: integer);
begin
  ftext.AddObject(AText,TObject(int(AWidth)));
end;

procedure TXSPagecontrol.AddPage(ACaption: string; refreshing: boolean);
var
  tablist: TStringList;
begin
  tablist := TStringList.Create;
  pagelist.AddObject(ACaption,tablist);
  factivepage := pagelist.count - 1;
  fstartpage  := fstartpage + pagelist.count - fendpage;
  InvisibleActiveTab;
  if refreshing then Refresh;
end;

function TXSPagecontrol.AddTab(ACaption, APath: string): TXSScrollbox;
var
  sb: TXSScrollbox;
begin
  if TStringList(pagelist.objects[factivepage]).count > fmaxtab -1 then exit;
  sb                     := TXSScrollbox.Create(self);
  sb.Parent              := self;
  sb.Caption             := ACaption;
  sb.Path                := APath;
  sb.Modified            := false;
  sb.ActiveTab           := true;
  if TStringList(pagelist.objects[factivepage]).count > 0 then begin
    sb.visible           := false;
    sb.ActiveTab         := false;
  end;
  sb.Align               := alClient;
  sb.Anchors             := [akTop,akLeft,akRight,akBottom];
  sb.BorderSpacing.Left  := frectoffset+2;
  sb.BorderSpacing.Top   := fpageheight+1;
  sb.BorderSpacing.Right := frectoffset+2;
  sb.BorderSpacing.Bottom:= ftabheight +1;
  TStringList(pagelist.objects[factivepage]).AddObject(ACaption,sb);
  result := sb;
end;

procedure TXSPagecontrol.DrawPage(acaption: string; r: TRect; flag,closed,modified: boolean);
const
  radius = 5;
var
  i: integer;
  e: TRect;
  p1,p2,p3,p4: TPoint;
  s  : string;
begin
  dec(r.left,fstartpage*fpagewidth);
  dec(r.right,fstartpage*fpagewidth);
  if flag then begin
    Canvas.Pen.Color   := fafrcolor;
    Canvas.Brush.Color := fabgcolor;
    Canvas.Font.Color  := fafgcolor;
    if fmousemove then Canvas.Font.Color:= clMaroon;
  end else begin
    Canvas.Pen.Color   := ffrcolor;
    Canvas.Brush.Color := fbgcolor;
    Canvas.Font.Color  := ffgcolor;
  end;

  e := Rect(r.left+fpageangle,r.top,r.left+radius*2+fpageangle,r.top+radius*2);
  Canvas.Ellipse(e);
  e := Rect(r.right-radius*2-fpageangle,r.top,r.right-fpageangle,r.top+radius*2);
  Canvas.Ellipse(e);

  p1 := Point(r.left-fpageangle,r.bottom-1);
  p2 := Point(r.left+fpageangle,r.top+radius);
  p3 := Point(r.right-fpageangle,r.top+radius);
  p4 := Point(r.right+fpageangle,r.bottom-1);
  Canvas.Polygon([p1,p2,p3,p4]);

  if flag then begin
    Canvas.Pen.Color   := fabgcolor;
    Canvas.Brush.Color := fabgcolor;
    Canvas.Line(r.left-fpageangle,r.bottom-1,r.right+fpageangle,r.bottom-1);
  end else begin
    Canvas.Pen.Color   := fbgcolor;
    Canvas.Brush.Color := fbgcolor;
  end;
  e := Rect(r.left+fpageangle+radius+1,r.top,r.right-fpageangle-radius-1,r.top+radius+1);
  Canvas.Rectangle(e);
  Canvas.Line(r.left+fpageangle+1,r.top+radius,r.right-fpageangle,r.top+radius);

  if flag then begin
    Canvas.Pen.Color   := fafrcolor;
  end else begin
    Canvas.Pen.Color   := ffrcolor;
  end;
  Canvas.Line(r.left+fpageangle+radius+1,r.top,r.right-fpageangle-radius-1,r.top);

  if closed then
    if flag then
      Canvas.Pen.Color  := clRed
    else
      Canvas.Pen.Color  := clGray;

  e := Rect(r.right-fpageangle*2-fclosewidth,r.top+(r.bottom-r.top-fclosewidth) div 2,r.right-fpageangle*2,r.top+(r.bottom-r.top+fclosewidth) div 2);
  Canvas.Line(e.Left+1,e.Top+1,e.Right,e.Bottom);
  Canvas.Line(e.Left+1,e.Top+2,e.Right-1,e.Bottom);
  Canvas.Line(e.Left+2,e.Top+1,e.Right,e.Bottom-1);
  Canvas.Line(e.Left+1,e.Bottom-1,e.Right,e.Top);
  Canvas.Line(e.Left+1,e.Bottom-2,e.Right-1,e.Top);
  Canvas.Line(e.Left+2,e.Bottom-1,e.Right,e.Top+1);

  s := ''; i := 1;
  while (Canvas.TextWidth(s+'....')<fpagewidth-fpageangle*4-fclosewidth) and (length(s)<length(acaption)) do begin
    s := s + acaption[i];
    inc(i);
  end;
  if (i<length(acaption)) then s:=s+'...';

  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(r.left+fpageangle*2,r.top+(r.bottom-r.top-Canvas.TextHeight(s)) div 2,s);

  if modified then begin
    Canvas.Pen.Color   := clMaroon;
    Canvas.Brush.Color := clMaroon;
    e := Rect(r.left+(r.right-r.left) div 2 - fpagewidth div 5,r.top+1,r.left+(r.right-r.left) div 2 + fpagewidth div 5,r.top+3);
    Canvas.Rectangle(e);
  end;

end;

procedure TXSPagecontrol.DrawTab(AText: string; r: TRect; flag: boolean);
const
  radius = 5;
var
  p1,p2,p3,p4: TPoint;
begin

  if flag then begin
    Canvas.Pen.Color   := fafrcolor;
    Canvas.Brush.Color := fabgcolor;
    Canvas.Font.Color  := fafgcolor;
  end else begin
    Canvas.Pen.Color   := ffrcolor;
    Canvas.Brush.Color := fbgcolor;
    Canvas.Font.Color  := ffgcolor;
  end;
  Canvas.Roundrect(r,radius,radius);

  p1 := Point(r.left-3,r.top);
  p2 := Point(r.left,r.top+3);
  p3 := Point(r.right,r.top+3);
  p4 := Point(r.right+3,r.top);
  Canvas.Polygon([p1,p2,p3,p4]);

  if flag then begin
    Canvas.Pen.Color   := fabgcolor;
    Canvas.Line(r.left-3+1,r.top,r.right+3,r.top);
  end else begin
    Canvas.Pen.Color   := fbgcolor;
  end;
  Canvas.Line(r.left+1,r.top+3,r.right,r.top+3);

  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(r.left+(r.right-r.left-Canvas.TextWidth(AText)) div 2,r.top+(r.bottom-r.top-Canvas.TextHeight(AText)) div 2,AText);
end;

procedure TXSPagecontrol.DrawText;
var
  i,j,k: integer;
begin
  j := 0;
  for i:=0 to ftext.count-1 do begin
    Canvas.Font.Color:= ffgcolor;
    Canvas.TextOut(fpageoffset+fmaxtab*ftabwidth+10+j,height-2-fpageheight + fpageheight div 2 + Canvas.Font.Height div 2 - 1,'|');
    Canvas.Font.Color:= fafgcolor;
    k := 5;
    if i<=2 then k := (int(ftext.Objects[i]) - Canvas.TextWidth(ftext[i])) div 2;
    Canvas.TextOut(fpageoffset+fmaxtab*ftabwidth+10+j+1+k,height-2-fpageheight + fpageheight div 2 + Canvas.Font.Height div 2,ftext[i]);
    j := j + int(ftext.Objects[i]);
  end;
end;

procedure TXSPagecontrol.Paint;
var
  up,down,i,a: integer;
  r,tr: TRect;
  page: TStringList;
  f   : boolean;
begin

  up := 0; down := 0;
  if pagelist.count>0 then up := fpageheight-frectoffset-1;
  if (factivepage>-1) and (TStringList(pagelist.objects[factivepage]).count>0) then down := ftabheight-frectoffset-1;
  fcanvasrect := Rect(frectoffset,frectoffset+up,width-frectoffset,height-frectoffset-down);
  flrrect     := Rect(width-frectoffset-flrwidth,4,width-frectoffset,fpageheight-3);
  fmaxvisible := (flrrect.left - frectoffset - fpageoffset) div fpagewidth;
  if fmaxvisible < pagelist.count then flrvisible := true else flrvisible := false;
  if fmaxvisible > pagelist.count - 1 then fendpage := pagelist.count else fendpage := fstartpage + fmaxvisible;
  if fendpage > pagelist.Count then fendpage := pagelist.Count;
  if (fstartpage > 0) and (fmaxvisible > pagelist.Count - fstartpage) then dec(fstartpage);
  if (fstartpage < 0) then fstartpage := 0;
  if (fstartpage = 0) and (pagelist.Count>=fmaxvisible) then fendpage := fmaxvisible;

  Canvas.Pen.Color   := ffrcolor;
  Canvas.Brush.Color := fabgcolor;
  Canvas.Roundrect(fcanvasrect,fcradius,fcradius);

  f:=false;
  for i := fstartpage to fendpage - 1 do begin
    r := Rect(fpageoffset+i*fpagewidth,1,fpageoffset+i*fpagewidth+fpagewidth,fpageheight);
    if i<>factivepage then
      DrawPage(pagelist[i],r,false,fcloseidx=i,TXSScrollBox(TStringList(pagelist.objects[i]).objects[0]).modified)
    else begin
      tr := r;
      f := (TStringList(pagelist.objects[i]).count>0) and TXSScrollBox(TStringList(pagelist.objects[i]).objects[0]).modified;
    end;
  end;
  if factivepage>=0 then begin
    DrawPage(pagelist[factivepage],tr,true,fcloseidx=factivepage,f);
    page := TStringList(pagelist.objects[factivepage]);
    a := 0;
    for i:=0 to page.count - 1 do begin
      r := Rect(ftaboffset+i*ftabwidth-1,height-ftabheight,ftaboffset+i*ftabwidth+ftabwidth,height-2);
      if not TXSScrollbox(page.objects[i]).activetab then
        DrawTab(page[i],r,false)
      else begin
        tr := r;
        a  := i;
      end;
    end;
    if page.count > 0 then DrawTab(page[a],tr,true);
    DrawText;
  end;

  if flrvisible then begin
    Canvas.Pen.Color   := ffrcolor;
    Canvas.Brush.Color := fbgcolor;
    Canvas.Font.Color  := ffgcolor;

    Canvas.RoundRect(flrrect,fpageheight-6,fpageheight-6);
    Canvas.Line(flrrect.Left+(flrrect.Right-flrrect.Left) div 2,flrrect.Top+3,flrrect.Left+(flrrect.Right-flrrect.Left) div 2,flrrect.Bottom-3);

    if flbactive then Canvas.Pen.Color  := clBlack else Canvas.Pen.Color  := ffrcolor;
    Canvas.Line(flrrect.Left+6,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Left+10,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 - 4);
    Canvas.Line(flrrect.Left+6,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Left+10,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 + 4);
    Canvas.Line(flrrect.Left+8,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Left+12,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 - 4);
    Canvas.Line(flrrect.Left+8,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Left+12,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 + 4);

    if flractive then Canvas.Pen.Color  := ffgcolor else Canvas.Pen.Color  := ffrcolor;
    Canvas.Line(flrrect.Right-6,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Right-10,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 - 4);
    Canvas.Line(flrrect.Right-6,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Right-10,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 + 4);
    Canvas.Line(flrrect.Right-8,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Right-12,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 - 4);
    Canvas.Line(flrrect.Right-8,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2,flrrect.Right-12,flrrect.Top+(flrrect.Bottom-flrrect.Top) div 2 + 4);
  end;

end;


end.

